"""
Variant QC related hail-query jobs
"""

import uuid
from os.path import join
from typing import List, Optional, Dict, Tuple
import logging
import hailtop.batch as hb
from hailtop.batch.job import Job

from analysis_runner import dataproc
from joint_calling import utils
from joint_calling.vqsr import make_vqsr_jobs

logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


def add_variant_qc_jobs(
    b: hb.Batch,
    work_bucket: str,
    web_bucket: str,
    raw_combined_mt_path: str,
    hard_filter_ht_path: str,
    meta_ht_path: str,
    sample_count: int,
    ped_file: Optional[str],
    overwrite: bool,
    vqsr_params_d: Dict,
    scatter_count: int,
    depends_on: Optional[List[Job]] = None,
    run_rf: bool = False,
) -> Tuple[Job, str]:
    """
    Add variant QC related hail-query jobs
    """
    rf_bucket = join(work_bucket, 'rf')
    vqsr_bucket = join(work_bucket, 'vqsr')

    fam_stats_ht_path = join(work_bucket, 'fam-stats.ht') if ped_file else None
    allele_data_ht_path = join(work_bucket, 'allele-data.ht')
    qc_ac_ht_path = join(work_bucket, 'qc-ac.ht')
    rf_result_ht_path = None

    info_ht_path = join(work_bucket, 'info.ht')
    info_split_ht_path = join(work_bucket, 'info-split.ht')
    if any(
        not utils.can_reuse(fp, overwrite) for fp in [info_ht_path, info_split_ht_path]
    ):
        info_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/generate_info_ht.py --overwrite '
            f'--mt {raw_combined_mt_path} '
            f'--out-info-ht {info_ht_path} '
            f'--out-split-info-ht {info_split_ht_path}',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=depends_on or [],
            job_name='Var QC: generate info',
        )
    else:
        info_job = b.new_job('Var QC: generate info [reuse]')

    if any(
        not utils.can_reuse(fp, overwrite)
        for fp in [allele_data_ht_path, qc_ac_ht_path]
    ):
        var_qc_anno_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/generate_variant_qc_annotations.py '
            + f'{"--overwrite " if overwrite else ""}'
            + f'--mt {raw_combined_mt_path} '
            + f'--hard-filtered-samples-ht {hard_filter_ht_path} '
            + f'--meta-ht {meta_ht_path} '
            + f'--out-allele-data-ht {allele_data_ht_path} '
            + f'--out-qc-ac-ht {qc_ac_ht_path} '
            + (f'--out-fam-stats-ht {fam_stats_ht_path} ' if ped_file else '')
            + (f'--fam-file {ped_file} ' if ped_file else '')
            + f'--bucket {work_bucket} '
            + f'--n-partitions {scatter_count * 25}',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=depends_on or [],
            job_name='Var QC: generate annotations',
            vep='GRCh38',
        )
    else:
        var_qc_anno_job = b.new_job('Var QC: generate annotations [reuse]')

    freq_ht_path = join(work_bucket, 'frequencies.ht')
    if overwrite or not utils.file_exists(freq_ht_path):
        freq_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/generate_freq_data.py --overwrite '
            f'--mt {raw_combined_mt_path} '
            f'--hard-filtered-samples-ht {hard_filter_ht_path} '
            f'--meta-ht {meta_ht_path} '
            f'--out-ht {freq_ht_path} '
            f'--bucket {work_bucket} ',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=depends_on or [],
            job_name='Var QC: generate frequencies',
        )
    else:
        freq_job = b.new_job('Var QC: generate frequencies [reuse]')

    rf_annotations_ht_path = join(work_bucket, 'rf-annotations.ht')
    if overwrite or not utils.file_exists(rf_annotations_ht_path):
        rf_anno_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/create_rf_annotations.py --overwrite '
            f'--info-split-ht {info_split_ht_path} '
            f'--freq-ht {freq_ht_path} '
            + (f'--fam-stats-ht {fam_stats_ht_path} ' if fam_stats_ht_path else '')
            + f'--allele-data-ht {allele_data_ht_path} '
            f'--qc-ac-ht {qc_ac_ht_path} '
            f'--bucket {work_bucket} '
            f'--use-adj-genotypes '
            f'--out-ht {rf_annotations_ht_path} '
            + f'--n-partitions {scatter_count * 25}',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=[freq_job, var_qc_anno_job, info_job],
            job_name='Var QC: create RF annotations',
        )
    else:
        rf_anno_job = b.new_job('Var QC: create RF annotations [reuse]')

    if run_rf:
        rf_result_ht_path = join(work_bucket, 'rf-result.ht')
        rf_model_id = f'rf_{str(uuid.uuid4())[:8]}'
        if overwrite or not utils.file_exists(rf_result_ht_path):
            rf_job = dataproc.hail_dataproc_job(
                b,
                f'{utils.SCRIPTS_DIR}/random_forest.py --overwrite '
                f'--annotations-ht {rf_annotations_ht_path} '
                f'--bucket {work_bucket} '
                f'--use-adj-genotypes '
                f'--out-results-ht {rf_result_ht_path} '
                f'--out-model-id {rf_model_id} ',
                max_age='8h',
                packages=utils.DATAPROC_PACKAGES,
                num_secondary_workers=scatter_count,
                depends_on=[rf_anno_job],
                job_name='Random forest',
            )
        else:
            rf_job = b.new_job('Random forest [reuse]')

        final_job, final_filter_ht_path = make_rf_eval_jobs(
            b=b,
            combined_mt_path=raw_combined_mt_path,
            info_split_ht_path=info_split_ht_path,
            rf_result_ht_path=rf_result_ht_path,
            rf_annotations_ht_path=rf_annotations_ht_path,
            fam_stats_ht_path=fam_stats_ht_path,
            freq_ht_path=freq_ht_path,
            rf_model_id=rf_model_id,
            work_bucket=rf_bucket,
            overwrite=overwrite,
            depends_on=[rf_job],
            scatter_count=scatter_count,
        )

    else:
        vqsred_vcf_path = join(vqsr_bucket, 'output.vcf.gz')
        if overwrite or not utils.file_exists(vqsred_vcf_path):
            final_gathered_vcf_job = make_vqsr_jobs(
                b,
                combined_mt_path=raw_combined_mt_path,
                hard_filter_ht_path=hard_filter_ht_path,
                meta_ht_path=meta_ht_path,
                gvcf_count=sample_count,
                work_bucket=vqsr_bucket,
                web_bucket=join(web_bucket, 'vqsr'),
                depends_on=depends_on or [],
                vqsr_params_d=vqsr_params_d,
                scatter_count=scatter_count,
                output_vcf_path=vqsred_vcf_path,
                overwrite=overwrite,
            )
        else:
            final_gathered_vcf_job = b.new_job('VQSR [reuse]')

        final_filter_ht_path = join(vqsr_bucket, 'final-filter.ht')
        final_job = make_vqsr_eval_jobs(
            b=b,
            combined_mt_path=raw_combined_mt_path,
            rf_annotations_ht_path=rf_annotations_ht_path,
            info_split_ht_path=info_split_ht_path,
            final_gathered_vcf_path=vqsred_vcf_path,
            rf_result_ht_path=rf_result_ht_path,
            fam_stats_ht_path=fam_stats_ht_path,
            freq_ht_path=freq_ht_path,
            work_bucket=vqsr_bucket,
            analysis_bucket=join(web_bucket, 'vqsr'),
            overwrite=overwrite,
            final_gathered_vcf_job=final_gathered_vcf_job,
            rf_anno_job=rf_anno_job,
            scatter_count=scatter_count,
            output_ht_path=final_filter_ht_path,
        )
    return final_job, final_filter_ht_path


def make_rf_eval_jobs(
    b: hb.Batch,
    combined_mt_path: str,
    info_split_ht_path: str,
    rf_result_ht_path: str,
    rf_annotations_ht_path: str,
    fam_stats_ht_path: Optional[str],
    freq_ht_path: str,
    rf_model_id: str,
    work_bucket: str,
    overwrite: bool,
    depends_on: Optional[List[Job]],
    scatter_count: int,
) -> Tuple[Job, str]:
    """
    Make jobs that do evaluation RF model and applies the final filters

    Returns the final_filter Job object and the path to the final filter HT
    """
    score_bin_ht_path = join(work_bucket, 'rf-score-bin.ht')
    score_bin_agg_ht_path = join(work_bucket, 'rf-score-agg-bin.ht')
    if overwrite or not utils.file_exists(score_bin_ht_path):
        eval_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/evaluation.py --overwrite '
            f'--mt {combined_mt_path} '
            f'--rf-annotations-ht {rf_annotations_ht_path} '
            f'--info-split-ht {info_split_ht_path} '
            + (f'--fam-stats-ht {fam_stats_ht_path} ' if fam_stats_ht_path else '')
            + f'--rf-results-ht {rf_result_ht_path} '
            f'--bucket {work_bucket} '
            f'--out-bin-ht {score_bin_ht_path} '
            f'--out-aggregated-bin-ht {score_bin_agg_ht_path} '
            f'--run-sanity-checks ',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=depends_on,
            job_name='RF: evaluation',
        )
    else:
        eval_job = b.new_job('RF: evaluation [reuse]')

    final_filter_ht_path = join(work_bucket, 'final-filter.ht')
    if overwrite or not utils.file_exists(final_filter_ht_path):
        final_filter_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/final_filter.py --overwrite '
            f'--out-final-filter-ht {final_filter_ht_path} '
            f'--model-id {rf_model_id} '
            f'--model-name RF '
            f'--score-name RF '
            f'--info-split-ht {info_split_ht_path} '
            f'--freq-ht {freq_ht_path} '
            f'--score-bin-ht {score_bin_ht_path} '
            f'--score-bin-agg-ht {score_bin_agg_ht_path} ' + f'--bucket {work_bucket} ',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=[eval_job],
            job_name='RF: final filter',
        )
    else:
        final_filter_job = b.new_job('RF: final filter [reuse]')

    return final_filter_job, final_filter_ht_path


def make_vqsr_eval_jobs(
    b: hb.Batch,
    combined_mt_path: str,
    rf_annotations_ht_path: str,
    info_split_ht_path: str,
    final_gathered_vcf_path: str,
    rf_result_ht_path: Optional[str],
    fam_stats_ht_path: Optional[str],
    freq_ht_path: str,
    work_bucket: str,
    analysis_bucket: str,  # pylint: disable=unused-argument
    overwrite: bool,
    final_gathered_vcf_job: Job,
    rf_anno_job: Job,
    scatter_count: int,
    output_ht_path: str,
) -> Tuple[Job, str]:
    """
    Make jobs that do evaluation VQSR model and applies the final filters

    Returns the final_filter Job object and the path to the final filter HT
    """
    vqsr_filters_split_ht_path = join(work_bucket, 'vqsr-filters-split.ht')
    if overwrite or not utils.file_exists(vqsr_filters_split_ht_path):
        load_vqsr_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/load_vqsr.py --overwrite '
            f'--split-multiallelic '
            f'--out-path {vqsr_filters_split_ht_path} '
            f'--vqsr-vcf-path {final_gathered_vcf_path} '
            f'--bucket {work_bucket} ',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=[final_gathered_vcf_job],
            job_name='VQSR: load_vqsr',
        )
    else:
        load_vqsr_job = b.new_job('VQSR: load_vqsr [reuse]')

    score_bin_ht_path = join(work_bucket, 'vqsr-score-bin.ht')
    score_bin_agg_ht_path = join(work_bucket, 'vqsr-score-agg-bin.ht')
    if (
        overwrite
        or not utils.file_exists(score_bin_ht_path)
        or not utils.file_exists(score_bin_agg_ht_path)
    ):
        eval_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/evaluation.py --overwrite '
            f'--mt {combined_mt_path} '
            f'--rf-annotations-ht {rf_annotations_ht_path} '
            f'--info-split-ht {info_split_ht_path} '
            + (f'--fam-stats-ht {fam_stats_ht_path} ' if fam_stats_ht_path else '')
            + (
                f'--rf-result-ht {rf_result_ht_path} '
                if (rf_annotations_ht_path and rf_result_ht_path)
                else ''
            )
            + f'--vqsr-filters-split-ht {vqsr_filters_split_ht_path} '
            f'--bucket {work_bucket} '
            f'--out-bin-ht {score_bin_ht_path} '
            f'--out-aggregated-bin-ht {score_bin_agg_ht_path} '
            f'--run-sanity-checks ',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=[load_vqsr_job, rf_anno_job],
            job_name='VQSR: evaluation',
        )
    else:
        eval_job = b.new_job('VQSR: evaluation [reuse]')

    vqsr_model_id = 'vqsr_model'
    if not utils.file_exists(output_ht_path):
        final_filter_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/final_filter.py --overwrite '
            f'--out-final-filter-ht {output_ht_path} '
            f'--vqsr-filters-split-ht {vqsr_filters_split_ht_path} '
            f'--model-id {vqsr_model_id} '
            f'--model-name VQSR '
            f'--score-name AS_VQSLOD '
            f'--info-split-ht {info_split_ht_path} '
            f'--freq-ht {freq_ht_path} '
            f'--score-bin-ht {score_bin_ht_path} '
            f'--score-bin-agg-ht {score_bin_agg_ht_path} '
            f'--bucket {work_bucket} ',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=[eval_job],
            job_name='VQSR: final filter',
        )
    else:
        final_filter_job = b.new_job('VQSR: final filter [reuse]')
    return final_filter_job



#!/usr/bin/env python3

"""
Functions that craete Batch jobs that get from raw data to a combiner-ready GVCF
"""

import logging
from os.path import join, dirname, splitext, basename
from typing import Optional, List, Tuple

import pandas as pd
import hailtop.batch as hb
from hailtop.batch.job import Job
from joint_calling import sm_utils
from joint_calling import utils


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


def add_pre_combiner_jobs(
    b: hb.Batch,
    samples_df: pd.DataFrame,
    pre_combiner_bucket: str,
    output_suffix: str,
    overwrite: bool,
    analysis_project: str,  # pylint: disable=unused-argument
) -> Tuple[pd.DataFrame, str, List[Job]]:
    """
    Add jobs that prepare GVCFs for the combiner, if needed.

    Returns a Tuple of: a DataFrame with the sample metadata, a TSV file
    corresponding to that dataframe, and a list of jobs to wait for before
    submitting the combiner job
    """

    logger.info(f'Samples DF: {samples_df}')
    jobs = []

    def get_project_bucket(_proj):
        if proj in ['syndip', 'giab']:
            proj_bucket = f'gs://cpg-reference/validation/{proj}'
        else:
            proj_bucket = f'gs://cpg-{proj}-{output_suffix}'
        return proj_bucket

    # Samples for which a GVCF is provided as input:
    gvcf_df = samples_df[samples_df.gvcf != '-']
    for s_id, proj, external_id, gvcf_path in zip(
        gvcf_df.s, gvcf_df.project, gvcf_df.external_id, gvcf_df.gvcf
    ):
        proj_bucket = get_project_bucket(proj)

        output_gvcf_path = join(proj_bucket, 'gvcf', f'{s_id}.g.vcf.gz')
        if not utils.can_reuse(output_gvcf_path, overwrite):
            j = _add_postproc_gvcf_jobs(
                b=b,
                gvcf_path=gvcf_path,
                output_path=output_gvcf_path,
                sample_name=s_id,
                project_name=proj,
                external_id=external_id,
                overwrite=overwrite,
            )
            jobs.append(j)
        samples_df.loc[s_id, ['gvcf']] = output_gvcf_path
        logger.info(f'Updating sample {s_id} gvcf to {output_gvcf_path}')

    logger.info(f'Updated sample DF after post-processing GVCF inputs: {samples_df}')

    # Samples for which a CRAM is provided as input:
    hc_intervals_j = None
    cram_df = samples_df[samples_df.cram != '-']
    for s_id, external_id, proj, input_cram, input_crai in zip(
        cram_df.s, cram_df.external_id, cram_df.project, cram_df.cram, cram_df.crai
    ):
        assert isinstance(input_crai, str)

        proj_bucket = get_project_bucket(proj)

        output_cram_fpath = join(proj_bucket, 'cram', f'{s_id}.cram')
        output_crai_fpath = join(proj_bucket, 'cram', f'{s_id}.cram.crai')
        if not utils.can_reuse(output_cram_fpath, overwrite):
            alignment_input = sm_utils.AlignmentInput(
                bam_or_cram_path=input_cram,
                index_path=input_crai,
            )
            cram_j = _make_realign_jobs(
                b=b,
                output_path=output_cram_fpath,
                sample_name=s_id,
                project_name=proj,
                alignment_input=alignment_input,
            )
        else:
            cram_j = None
        samples_df.loc[s_id, ['cram']] = output_cram_fpath
        samples_df.loc[s_id, ['crai']] = output_crai_fpath

        output_gvcf_path = join(proj_bucket, 'gvcf', f'{s_id}.g.vcf.gz')
        if not utils.can_reuse(output_gvcf_path, overwrite):
            if hc_intervals_j is None:
                hc_intervals_j = _add_split_intervals_job(
                    b=b,
                    interval_list=utils.UNPADDED_INTERVALS,
                    scatter_count=utils.NUMBER_OF_HAPLOTYPE_CALLER_INTERVALS,
                    ref_fasta=utils.REF_FASTA,
                )
            gvcf_j = _make_produce_gvcf_jobs(
                b=b,
                output_path=output_gvcf_path,
                sample_name=s_id,
                project_name=proj,
                external_id=external_id,
                cram_fpath=output_cram_fpath,
                crai_fpath=output_crai_fpath,
                intervals_j=hc_intervals_j,
                tmp_bucket=join(pre_combiner_bucket, 'tmp'),
                overwrite=overwrite,
                depends_on=[cram_j] if cram_j else [],
            )
            jobs.append(gvcf_j)
        samples_df.loc[s_id, ['gvcf']] = output_gvcf_path

    # Saving the resulting DataFrame as a TSV file
    gvcfs_tsv_path = join(pre_combiner_bucket, 'gvcfs.tsv')
    samples_df.to_csv(gvcfs_tsv_path, index=False, sep='\t', na_rep='NA')
    logger.info(f'Saved combiner-ready GVCF data to {gvcfs_tsv_path}')
    return samples_df, gvcfs_tsv_path, jobs


def _make_realign_jobs(
    b: hb.Batch,
    output_path: str,
    sample_name: str,
    project_name: str,
    alignment_input: sm_utils.AlignmentInput,
) -> Job:
    """
    Runs BWA to realign reads back again to hg38.

    When the input is CRAM/BAM, uses Bazam to stream reads to BWA.
    """
    job_name = f'{project_name}/{sample_name}: BWA align'
    if utils.file_exists(output_path):
        return b.new_job(f'{job_name} [reuse]')

    logger.info(
        f'Parsing the "reads" metadata field and submitting the alignment '
        f'to write {output_path} for {sample_name}. '
    )
    j = b.new_job(job_name)
    j.image(utils.ALIGNMENT_IMAGE)
    total_cpu = 32
    j.cpu(total_cpu)
    j.memory('standard')
    j.storage('500G')
    j.declare_resource_group(
        output_cram={
            'cram': '{root}.cram',
            'cram.crai': '{root}.cram.crai',
        }
    )

    reference = b.read_input_group(
        base=utils.REF_FASTA,
        fai=utils.REF_FASTA + '.fai',
        dict=utils.REF_FASTA.replace('.fasta', '')
        .replace('.fna', '')
        .replace('.fa', '')
        + '.dict',
        amb=utils.REF_FASTA + '.amb',
        ann=utils.REF_FASTA + '.ann',
        pac=utils.REF_FASTA + '.pac',
        o123=utils.REF_FASTA + '.0123',
        bwa2bit64=utils.REF_FASTA + '.bwt.2bit.64',
    )

    work_dir = dirname(j.output_cram.cram)

    pull_inputs_cmd = ''

    if alignment_input.bam_or_cram_path:
        use_bazam = True
        bazam_cpu = 10
        bwa_cpu = 32
        bamsormadup_cpu = 10
        assert alignment_input.index_path
        assert not alignment_input.fqs1 and not alignment_input.fqs2

        if alignment_input.bam_or_cram_path.startswith('gs://'):
            cram = b.read_input_group(
                base=alignment_input.bam_or_cram_path, index=alignment_input.index_path
            )
            cram_localized_path = cram.base
        else:
            # Can't use on Batch localization mechanism with `b.read_input_group`,
            # but have to manually localize with `wget`
            cram_name = basename(alignment_input.bam_or_cram_path)
            cram_localized_path = join(work_dir, cram_name)
            crai_localized_path = join(work_dir, cram_name + '.crai')
            pull_inputs_cmd = (
                f'wget {alignment_input.bam_or_cram_path} -O {cram_localized_path}\n'
                f'wget {alignment_input.index_path} -O {crai_localized_path}'
            )
        r1_param = (
            f'<(bazam -Xmx16g -Dsamjdk.reference_fasta={reference.base}'
            f' -n{bazam_cpu} -bam {cram_localized_path})'
        )
        r2_param = '-'
    else:
        assert alignment_input.fqs1 and alignment_input.fqs2
        use_bazam = False
        bwa_cpu = 32
        bamsormadup_cpu = 10
        files1 = [b.read_input(f1) for f1 in alignment_input.fqs1]
        files2 = [b.read_input(f1) for f1 in alignment_input.fqs2]
        r1_param = f'<(cat {" ".join(files1)})'
        r2_param = f'<(cat {" ".join(files2)})'
        logger.info(f'r1_param: {r1_param}')
        logger.info(f'r2_param: {r2_param}')

    rg_line = f'@RG\\tID:{sample_name}\\tSM:{sample_name}'
    # BWA command options:
    # -K     process INT input bases in each batch regardless of nThreads (for reproducibility)
    # -p     smart pairing (ignoring in2.fq)
    # -t16   threads
    # -Y     use soft clipping for supplementary alignments
    # -R     read group header line such as '@RG\tID:foo\tSM:bar'
    command = f"""
set -o pipefail
set -ex

(while true; do df -h; pwd; du -sh {work_dir}; sleep 600; done) &

{pull_inputs_cmd}

bwa-mem2 mem -K 100000000 {'-p' if use_bazam else ''} -t{bwa_cpu} -Y \\
    -R '{rg_line}' {reference.base} {r1_param} {r2_param} | \\
bamsormadup inputformat=sam threads={bamsormadup_cpu} SO=coordinate \\
    M={j.duplicate_metrics} outputformat=sam \\
    tmpfile=$(dirname {j.output_cram.cram})/bamsormadup-tmp | \\
samtools view -T {reference.base} -O cram -o {j.output_cram.cram}

samtools index -@{total_cpu} {j.output_cram.cram} {j.output_cram.crai}

df -h; pwd; du -sh {work_dir}
    """
    j.command(command)
    b.write_output(j.output_cram, splitext(output_path)[0])
    b.write_output(
        j.duplicate_metrics,
        join(
            dirname(output_path),
            'duplicate-metrics',
            f'{sample_name}-duplicate-metrics.csv',
        ),
    )
    return j


def _make_produce_gvcf_jobs(
    b: hb.Batch,
    output_path: str,
    sample_name: str,
    project_name: str,
    external_id: str,
    cram_fpath: str,
    crai_fpath: str,
    intervals_j: Job,
    tmp_bucket: str,
    overwrite: bool,
    depends_on: Optional[List[Job]] = None,
) -> Job:
    """
    Takes all samples with a 'file' of 'type'='bam' in `samples_df`,
    and runs HaplotypeCaller on them in parallel, then post-processes the GVCF

    HaplotypeCaller is run in an interval-based sharded way, with per-interval
    HaplotypeCaller jobs defined in a nested loop.
    """
    job_name = f'{project_name}/{sample_name}: make GVCF'
    if utils.file_exists(output_path):
        return b.new_job(f'{job_name} [reuse]')

    logger.info(
        f'Submitting the variant calling jobs to write {output_path} for {sample_name}'
    )

    reference = b.read_input_group(
        base=utils.REF_FASTA,
        fai=utils.REF_FASTA + '.fai',
        dict=utils.REF_FASTA.replace('.fasta', '')
        .replace('.fna', '')
        .replace('.fa', '')
        + '.dict',
    )

    haplotype_caller_jobs = []
    for idx in range(utils.NUMBER_OF_HAPLOTYPE_CALLER_INTERVALS):
        haplotype_caller_jobs.append(
            _add_haplotype_caller_job(
                b,
                sample_name=sample_name,
                project_name=project_name,
                cram_fpath=cram_fpath,
                crai_fpath=crai_fpath,
                interval=intervals_j.intervals[f'interval_{idx}'],
                reference=reference,
                interval_idx=idx,
                number_of_intervals=utils.NUMBER_OF_HAPLOTYPE_CALLER_INTERVALS,
                depends_on=depends_on,
            )
        )
    if depends_on:
        haplotype_caller_jobs[0].depends_on(*depends_on)

    hc_gvcf_path = join(tmp_bucket, 'haplotypecaller', f'{sample_name}.g.vcf.gz')
    merge_j = _add_merge_gvcfs_job(
        b=b,
        sample_name=sample_name,
        project_name=project_name,
        gvcfs=[j.output_gvcf for j in haplotype_caller_jobs],
        output_gvcf_path=hc_gvcf_path,
    )

    j = _add_postproc_gvcf_jobs(
        b=b,
        gvcf_path=hc_gvcf_path,
        output_path=output_path,
        sample_name=sample_name,
        project_name=project_name,
        external_id=external_id,
        overwrite=overwrite,
        depends_on=[merge_j],
    )
    return j


def _add_split_intervals_job(
    b: hb.Batch,
    interval_list: str,
    scatter_count: int,
    ref_fasta: str,
) -> Job:
    """
    Split genome into intervals to parallelise GnarlyGenotyper.

    Returns: a Job object with a single output j.intervals of type ResourceGroup
    """
    j = b.new_job(f'Make {scatter_count} intervals')
    j.image(utils.GATK_IMAGE)
    java_mem = 3
    j.memory('standard')  # ~ 4G/core ~ 4G
    j.storage('16G')
    j.declare_resource_group(
        intervals={
            f'interval_{idx}': f'{{root}}/{str(idx).zfill(4)}-scattered.interval_list'
            for idx in range(scatter_count)
        }
    )

    j.command(
        f"""set -e

    # Modes other than INTERVAL_SUBDIVISION will produce an unpredicted number 
    # of intervals. But we have to expect exactly the {scatter_count} number of 
    # output files because our workflow is not dynamic.
    gatk --java-options -Xms{java_mem}g SplitIntervals \\
      -L {interval_list} \\
      -O {j.intervals} \\
      -scatter {scatter_count} \\
      -R {ref_fasta} \\
      -mode INTERVAL_SUBDIVISION
      """
    )
    # Could save intervals to a bucket here to avoid rerunning the job
    return j


def _add_haplotype_caller_job(
    b: hb.Batch,
    sample_name: str,
    project_name: str,
    cram_fpath: str,
    crai_fpath: str,
    interval: hb.ResourceFile,
    reference: hb.ResourceGroup,
    interval_idx: Optional[int] = None,
    number_of_intervals: int = 1,
    depends_on: Optional[List[Job]] = None,
    output_gvcf_path: Optional[str] = None,
    overwrite: bool = False,
) -> Job:
    """
    Run HaplotypeCaller on an input BAM or CRAM, and output GVCF
    """
    job_name = f'{project_name}/{sample_name}: HaplotypeCaller'
    if interval_idx is not None:
        job_name += f', {sample_name} {interval_idx}/{number_of_intervals}'
    if utils.can_reuse(output_gvcf_path, overwrite):
        return b.new_job(f'{job_name} [reuse]')

    j = b.new_job(job_name)
    j.image(utils.GATK_IMAGE)
    j.cpu(2)
    java_mem = 7
    j.memory('standard')  # ~ 4G/core ~ 7.5G
    j.storage('60G')
    j.declare_resource_group(
        output_gvcf={
            'g.vcf.gz': '{root}-' + sample_name + '.g.vcf.gz',
            'g.vcf.gz.tbi': '{root}-' + sample_name + '.g.vcf.gz.tbi',
        }
    )
    if depends_on:
        j.depends_on(*depends_on)

    j.command(
        f"""set -e
    (while true; do df -h; pwd; du -sh $(dirname {j.output_gvcf['g.vcf.gz']}); free -m; sleep 300; done) &

    export GOOGLE_APPLICATION_CREDENTIALS=/gsa-key/key.json

    gatk --java-options "-Xms{java_mem}g -XX:GCTimeLimit=50 -XX:GCHeapFreeLimit=10" \\
      HaplotypeCaller \\
      -R {reference.base} \\
      -I {cram_fpath} \\
      --read-index {crai_fpath} \\
      -L {interval} \\
      -O {j.output_gvcf['g.vcf.gz']} \\
      -G AS_StandardAnnotation \\
      -GQB 20 \
      -ERC GVCF \\

    df -h; pwd; du -sh $(dirname {j.output_gvcf['g.vcf.gz']}); free -m
    """
    )
    if output_gvcf_path:
        b.write_output(j.output_gvcf, output_gvcf_path)
    return j


def _add_merge_gvcfs_job(
    b: hb.Batch,
    sample_name: str,
    project_name: str,
    gvcfs: List[hb.ResourceGroup],
    output_gvcf_path: Optional[str],
) -> Job:
    """
    Combine by-interval GVCFs into a single sample GVCF file
    """

    job_name = f'{project_name}/{sample_name}: merge {len(gvcfs)} GVCFs'
    j = b.new_job(job_name)
    j.image(utils.PICARD_IMAGE)
    j.cpu(2)
    java_mem = 7
    j.memory('standard')  # ~ 4G/core ~ 7.5G
    j.storage(f'{len(gvcfs) * 1.5 + 2}G')
    j.declare_resource_group(
        output_gvcf={
            'g.vcf.gz': '{root}-' + sample_name + '.g.vcf.gz',
            'g.vcf.gz.tbi': '{root}-' + sample_name + '.g.vcf.gz.tbi',
        }
    )

    input_cmd = ' '.join(f'INPUT={g["g.vcf.gz"]}' for g in gvcfs)

    j.command(
        f"""set -e

    (while true; do df -h; pwd; du -sh $(dirname {j.output_gvcf['g.vcf.gz']}); free -m; sleep 300; done) &

    java -Xms{java_mem}g -jar /usr/picard/picard.jar \
      MergeVcfs {input_cmd} OUTPUT={j.output_gvcf['g.vcf.gz']}

    df -h; pwd; du -sh $(dirname {j.output_gvcf['g.vcf.gz']}); free -m
      """
    )
    if output_gvcf_path:
        b.write_output(j.output_gvcf, output_gvcf_path.replace('.g.vcf.gz', ''))
    return j


def _add_postproc_gvcf_jobs(
    b: hb.Batch,
    gvcf_path: str,
    output_path: str,
    sample_name: str,
    project_name: str,
    external_id: str,
    overwrite: bool,
    depends_on: Optional[List[Job]] = None,
) -> Job:
    if utils.can_reuse(output_path, overwrite):
        return output_path
    logger.info(
        f'Adding reblock and subset jobs for sample {sample_name}, gvcf {gvcf_path}'
    )
    gvcf = b.read_input_group(
        **{'g.vcf.gz': gvcf_path, 'g.vcf.gz.tbi': gvcf_path + '.tbi'}
    )
    reblock_j = _add_reblock_gvcf_job(
        b=b,
        sample_name=sample_name,
        project_name=project_name,
        input_gvcf=gvcf,
        overwrite=overwrite,
    )
    if depends_on:
        reblock_j.depends_on(*depends_on)
    j = _add_subset_noalt_step(
        b=b,
        sample_name=sample_name,
        project_name=project_name,
        input_gvcf=reblock_j.output_gvcf,
        output_gvcf_path=output_path,
        noalt_regions=b.read_input(utils.NOALT_REGIONS),
        external_sample_id=external_id,
        internal_sample_id=sample_name,
        depends_on=depends_on,
        overwrite=overwrite,
    )
    return j


def _add_reblock_gvcf_job(
    b: hb.Batch,
    sample_name: str,
    project_name: str,
    input_gvcf: hb.ResourceGroup,
    overwrite: bool,
    output_gvcf_path: Optional[str] = None,
) -> Job:
    """
    Runs ReblockGVCF to annotate with allele-specific VCF INFO fields
    required for recalibration
    """
    job_name = f'{project_name}/{sample_name}: ReblockGVCF'
    if utils.can_reuse(output_gvcf_path, overwrite):
        return b.new_job(job_name + ' [reuse]')

    j = b.new_job(job_name)
    j.image(utils.GATK_IMAGE)
    mem_gb = 8
    j.memory(f'{mem_gb}G')
    j.storage(f'30G')
    j.declare_resource_group(
        output_gvcf={
            'g.vcf.gz': '{root}.g.vcf.gz',
            'g.vcf.gz.tbi': '{root}.g.vcf.gz.tbi',
        }
    )

    j.command(
        f"""
    gatk --java-options "-Xms{mem_gb - 1}g" \\
        ReblockGVCF \\
        -V {input_gvcf['g.vcf.gz']} \\
        -do-qual-approx \\
        -O {j.output_gvcf['g.vcf.gz']} \\
        --create-output-variant-index true"""
    )
    if output_gvcf_path:
        b.write_output(j.output_gvcf, output_gvcf_path.replace('.g.vcf.gz', ''))
    return j


def _add_subset_noalt_step(
    b: hb.Batch,
    sample_name: str,
    project_name: str,
    input_gvcf: hb.ResourceGroup,
    output_gvcf_path: str,
    noalt_regions: hb.ResourceFile,
    external_sample_id: str,
    internal_sample_id: str,
    overwrite: bool,
    depends_on: Optional[List[Job]] = None,
) -> Job:
    """
    1. Subset GVCF to main chromosomes to avoid downstream errors
    2. Removes the DS INFO field that is added to some HGDP GVCFs to avoid errors
       from Hail about mismatched INFO annotations
    3. Renames sample name from external_sample_id to internal_sample_id
    """
    job_name = f'{project_name}/{sample_name}: SubsetToNoalt'
    if utils.can_reuse(output_gvcf_path, overwrite):
        return b.new_job(job_name + ' [reuse]')

    j = b.new_job(job_name)
    j.image(utils.BCFTOOLS_IMAGE)
    mem_gb = 8
    j.memory(f'{mem_gb}G')
    j.storage(f'30G')
    j.declare_resource_group(
        output_gvcf={
            'g.vcf.gz': '{root}.g.vcf.gz',
            'g.vcf.gz.tbi': '{root}.g.vcf.gz.tbi',
        }
    )
    if depends_on:
        j.depends_on(*depends_on)
    j.command(
        f"""set -e

    bcftools view {input_gvcf['g.vcf.gz']} -T {noalt_regions} \\
        | bcftools annotate -x INFO/DS \\
        | bcftools reheader -s <(echo "{external_sample_id} {internal_sample_id}") \\
        | bcftools view -Oz -o {j.output_gvcf['g.vcf.gz']}

    bcftools index --tbi {j.output_gvcf['g.vcf.gz']}
        """
    )
    b.write_output(j.output_gvcf, output_gvcf_path.replace('.g.vcf.gz', ''))
    return j



"""
Create jobs to create and apply a VQSR model
"""

import os
from os.path import join
from typing import List, Optional, Dict
import logging
import hailtop.batch as hb
from hailtop.batch.job import Job
from analysis_runner import dataproc

from joint_calling import utils
from joint_calling.utils import can_reuse

logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


SNP_RECALIBRATION_TRANCHE_VALUES = [
    100.0,
    99.95,
    99.9,
    99.8,
    99.6,
    99.5,
    99.4,
    99.3,
    99.0,
    98.0,
    97.0,
    90.0,
]
SNP_RECALIBRATION_ANNOTATION_VALUES = [
    'AS_QD',
    'AS_MQRankSum',
    'AS_ReadPosRankSum',
    'AS_FS',
    'AS_SOR',
    'AS_MQ',
]
INDEL_RECALIBRATION_TRANCHE_VALUES = [
    100.0,
    99.95,
    99.9,
    99.5,
    99.0,
    97.0,
    96.0,
    95.0,
    94.0,
    93.5,
    93.0,
    92.0,
    91.0,
    90.0,
]
INDEL_RECALIBRATION_ANNOTATION_VALUES = [
    'AS_FS',
    'AS_SOR',
    'AS_ReadPosRankSum',
    'AS_MQRankSum',
    'AS_QD',
]


def make_vqsr_jobs(
    b: hb.Batch,
    combined_mt_path: str,
    hard_filter_ht_path: str,
    meta_ht_path: str,
    gvcf_count: int,
    work_bucket: str,
    web_bucket: str,
    depends_on: Optional[List[Job]],
    vqsr_params_d: Dict,
    scatter_count: int,
    output_vcf_path: str,
    overwrite: bool,
) -> Job:
    """
    Add jobs that perform the allele-specific VQSR variant QC

    :param b: Batch object to add jobs to
    :param combined_mt_path: path to a Matrix Table combined with the Hail VCF combiner
    :param hard_filter_ht_path: path to HT with samples that failed QC
    :param meta_ht_path: path to HT with sample QC metadata
    :param gvcf_count: number of input samples. Can't read from combined_mt_path as it
           might not be yet genereated the point of Batch job submission
    :param work_bucket: bucket for intermediate files
    :param web_bucket: bucket for plots and evaluation results (exposed via http)
    :param depends_on: job that the created jobs should only run after
    :param vqsr_params_d: parameters for VQSR
    :param scatter_count: number of shards to patition data for scattering
    :param output_vcf_path: path to write final recalibrated VCF to
    :param overwrite: whether to not reuse existing intermediate and output files
    :return: a final Job, and a path to the VCF with VQSR annotations
    """

    # Reference files. All options have defaults.
    unpadded_intervals_path = os.path.join(
        utils.REF_BUCKET, 'hg38.even.handcurated.20k.intervals'
    )
    dbsnp_vcf = os.path.join(utils.REF_BUCKET, 'Homo_sapiens_assembly38.dbsnp138.vcf')
    dbsnp_vcf_index = os.path.join(
        utils.REF_BUCKET, 'Homo_sapiens_assembly38.dbsnp138.vcf.idx'
    )
    eval_interval_list = os.path.join(
        utils.REF_BUCKET, 'wgs_evaluation_regions.hg38.interval_list'
    )
    hapmap_resource_vcf = os.path.join(utils.REF_BUCKET, 'hapmap_3.3.hg38.vcf.gz')
    hapmap_resource_vcf_index = os.path.join(
        utils.REF_BUCKET, 'hapmap_3.3.hg38.vcf.gz.tbi'
    )
    omni_resource_vcf = os.path.join(utils.REF_BUCKET, '1000G_omni2.5.hg38.vcf.gz')
    omni_resource_vcf_index = os.path.join(
        utils.REF_BUCKET, '1000G_omni2.5.hg38.vcf.gz.tbi'
    )
    one_thousand_genomes_resource_vcf = os.path.join(
        utils.REF_BUCKET, '1000G_phase1.snps.high_confidence.hg38.vcf.gz'
    )
    one_thousand_genomes_resource_vcf_index = os.path.join(
        utils.REF_BUCKET, '1000G_phase1.snps.high_confidence.hg38.vcf.gz.tbi'
    )
    mills_resource_vcf = os.path.join(
        utils.REF_BUCKET, 'Mills_and_1000G_gold_standard.indels.hg38.vcf.gz'
    )
    mills_resource_vcf_index = os.path.join(
        utils.REF_BUCKET, 'Mills_and_1000G_gold_standard.indels.hg38.vcf.gz.tbi'
    )
    axiom_poly_resource_vcf = os.path.join(
        utils.REF_BUCKET, 'Axiom_Exome_Plus.genotypes.all_populations.poly.hg38.vcf.gz'
    )
    axiom_poly_resource_vcf_index = os.path.join(
        utils.REF_BUCKET,
        'Axiom_Exome_Plus.genotypes.all_populations.poly.hg38.vcf.gz.tbi',
    )
    ref_fasta = utils.REF_FASTA
    ref_fasta_index = utils.REF_FASTA + '.fai'
    ref_dict = (
        ref_fasta.replace('.fasta', '').replace('.fna', '').replace('.fa', '') + '.dict'
    )
    ref_fasta = b.read_input_group(
        base=ref_fasta,
        dict=ref_dict,
        fai=ref_fasta_index,
    )
    dbsnp_vcf = b.read_input_group(base=dbsnp_vcf, index=dbsnp_vcf_index)
    eval_interval_list = b.read_input(eval_interval_list)
    hapmap_resource_vcf = b.read_input_group(
        base=hapmap_resource_vcf, index=hapmap_resource_vcf_index
    )
    omni_resource_vcf = b.read_input_group(
        base=omni_resource_vcf, index=omni_resource_vcf_index
    )
    one_thousand_genomes_resource_vcf = b.read_input_group(
        base=one_thousand_genomes_resource_vcf,
        index=one_thousand_genomes_resource_vcf_index,
    )
    mills_resource_vcf = b.read_input_group(
        base=mills_resource_vcf, index=mills_resource_vcf_index
    )
    axiom_poly_resource_vcf = b.read_input_group(
        base=axiom_poly_resource_vcf, index=axiom_poly_resource_vcf_index
    )
    dbsnp_resource_vcf = dbsnp_vcf

    is_small_callset = gvcf_count < 1000
    # 1. For small callsets, we don't apply the ExcessHet filtering.
    # 2. For small callsets, we gather the VCF shards and collect QC metrics directly.
    # For anything larger, we need to keep the VCF sharded and gather metrics
    # collected from them.
    is_huge_callset = gvcf_count >= 100000
    # For huge callsets, we allocate more memory for the SNPs Create Model step

    small_disk = 30 if is_small_callset else (50 if not is_huge_callset else 100)
    medium_disk = 50 if is_small_callset else (100 if not is_huge_callset else 200)
    huge_disk = 100 if is_small_callset else (500 if not is_huge_callset else 2000)

    combined_vcf_path = join(work_bucket, 'input.vcf.gz')
    if not can_reuse(combined_vcf_path, overwrite):
        mt_to_vcf_job = dataproc.hail_dataproc_job(
            b,
            f'{utils.SCRIPTS_DIR}/mt_to_vcf.py --overwrite '
            f'--mt {combined_mt_path} '
            f'--meta-ht {meta_ht_path} '
            f'--hard-filtered-samples-ht {hard_filter_ht_path} '
            f'-o {combined_vcf_path} ',
            max_age='8h',
            packages=utils.DATAPROC_PACKAGES,
            num_secondary_workers=scatter_count,
            depends_on=depends_on,
            # hl.export_vcf() uses non-preemptible workers' disk to merge VCF files.
            # 10 samples take 2.3G, 400 samples take 60G, which roughly matches
            # `huge_disk` (also used in the AS-VQSR VCF-gather job)
            worker_boot_disk_size=huge_disk,
            job_name='MT to VCF',
        )
    else:
        mt_to_vcf_job = b.new_job('VQSR: MT to VCF [reuse]')

    split_intervals_job = add_split_intervals_step(
        b,
        unpadded_intervals_path,
        scatter_count,
        ref_fasta,
        disk_size=small_disk,
    )
    intervals = split_intervals_job.intervals

    tabix_job = add_tabix_step(b, combined_vcf_path, medium_disk)
    tabix_job.depends_on(mt_to_vcf_job)

    gathered_vcf = tabix_job.combined_vcf
    scattered_vcfs = [gathered_vcf for _ in range(scatter_count)]

    if not is_small_callset:
        # ExcessHet filtering applies only to callsets with a large number of samples,
        # e.g. hundreds of unrelated samples. Small cohorts should not trigger ExcessHet
        # filtering as values should remain small. Note cohorts of consanguinous samples
        # will inflate ExcessHet, and it is possible to limit the annotation to founders
        # for such cohorts by providing a pedigree file during variant calling.
        hard_filtered_vcfs = [
            add_hard_filter_step(
                b,
                input_vcf=gathered_vcf,
                interval=intervals[f'interval_{idx}'],
                excess_het_threshold=vqsr_params_d['min_excess_het'],
                disk_size=medium_disk,
            ).output_vcf
            for idx in range(scatter_count)
        ]
        scattered_vcfs = hard_filtered_vcfs

        gathered_vcf = add_sites_only_gather_vcf_step(
            b,
            input_vcfs=scattered_vcfs,
            disk_size=medium_disk,
        ).output_vcf

    indels_variant_recalibrator_job = add_indels_variant_recalibrator_step(
        b,
        sites_only_variant_filtered_vcf=gathered_vcf,
        mills_resource_vcf=mills_resource_vcf,
        axiom_poly_resource_vcf=axiom_poly_resource_vcf,
        dbsnp_resource_vcf=dbsnp_resource_vcf,
        disk_size=small_disk,
        web_bucket=web_bucket,
        work_bucket=web_bucket,
    )
    indels_recalibration = indels_variant_recalibrator_job.recalibration
    indels_tranches = indels_variant_recalibrator_job.tranches

    snp_max_gaussians = 6
    if is_small_callset:
        snp_max_gaussians = 4
    elif is_huge_callset:
        snp_max_gaussians = 8

    if is_huge_callset:
        # Run SNP recalibrator in a scattered mode
        model_file = add_snps_variant_recalibrator_create_model_step(
            b,
            sites_only_variant_filtered_vcf=gathered_vcf,
            hapmap_resource_vcf=hapmap_resource_vcf,
            omni_resource_vcf=omni_resource_vcf,
            one_thousand_genomes_resource_vcf=one_thousand_genomes_resource_vcf,
            dbsnp_resource_vcf=dbsnp_resource_vcf,
            disk_size=small_disk,
            web_bucket=web_bucket,
            work_bucket=work_bucket,
            is_small_callset=is_small_callset,
            is_huge_callset=is_huge_callset,
            max_gaussians=snp_max_gaussians,
        ).model_file
        # model_file = b.read_input('gs://playground-au/batch/859e9a/18/model_report')

        snps_recalibrator_jobs = [
            add_snps_variant_recalibrator_scattered_step(
                b,
                sites_only_vcf=scattered_vcfs[idx],
                interval=intervals[f'interval_{idx}'],
                model_file=model_file,
                hapmap_resource_vcf=hapmap_resource_vcf,
                omni_resource_vcf=omni_resource_vcf,
                one_thousand_genomes_resource_vcf=one_thousand_genomes_resource_vcf,
                dbsnp_resource_vcf=dbsnp_resource_vcf,
                disk_size=small_disk,
                max_gaussians=snp_max_gaussians,
            )
            for idx in range(scatter_count)
        ]
        snps_recalibrations = [j.recalibration for j in snps_recalibrator_jobs]
        snps_tranches = [j.tranches for j in snps_recalibrator_jobs]
        snps_gathered_tranches = add_snps_gather_tranches_step(
            b,
            tranches=snps_tranches,
            disk_size=small_disk,
        ).out_tranches

        scattered_vcfs = [
            add_apply_recalibration_step(
                b,
                input_vcf=scattered_vcfs[idx],
                interval=intervals[f'interval_{idx}'],
                indels_recalibration=indels_recalibration,
                indels_tranches=indels_tranches,
                snps_recalibration=snps_recalibrations[idx],
                snps_tranches=snps_gathered_tranches,
                disk_size=huge_disk,
                indel_filter_level=vqsr_params_d['indel_filter_level'],
                snp_filter_level=vqsr_params_d['snp_filter_level'],
            ).recalibrated_vcf
            for idx in range(scatter_count)
        ]
        recalibrated_gathered_vcf_job = _add_final_gather_vcf_step(
            b,
            input_vcfs=scattered_vcfs,
            disk_size=huge_disk,
        )
        recalibrated_gathered_vcf = recalibrated_gathered_vcf_job.output_vcf

    else:
        snps_recalibrator_job = add_snps_variant_recalibrator_step(
            b,
            sites_only_variant_filtered_vcf=gathered_vcf,
            hapmap_resource_vcf=hapmap_resource_vcf,
            omni_resource_vcf=omni_resource_vcf,
            one_thousand_genomes_resource_vcf=one_thousand_genomes_resource_vcf,
            dbsnp_resource_vcf=dbsnp_resource_vcf,
            disk_size=small_disk,
            max_gaussians=snp_max_gaussians,
            web_bucket=web_bucket,
            work_bucket=work_bucket,
        )
        snps_recalibration = snps_recalibrator_job.recalibration
        snps_tranches = snps_recalibrator_job.tranches

        recalibrated_gathered_vcf_job = add_apply_recalibration_step(
            b,
            input_vcf=gathered_vcf,
            indels_recalibration=indels_recalibration,
            indels_tranches=indels_tranches,
            snps_recalibration=snps_recalibration,
            snps_tranches=snps_tranches,
            disk_size=huge_disk,
            indel_filter_level=vqsr_params_d['indel_filter_level'],
            snp_filter_level=vqsr_params_d['snp_filter_level'],
        )
        recalibrated_gathered_vcf = recalibrated_gathered_vcf_job.recalibrated_vcf

    final_gathered_vcf_job = _add_filter_sb_step(
        b,
        input_vcf=recalibrated_gathered_vcf,
        disk_size=medium_disk,
        output_vcf_path=output_vcf_path,
    )

    _add_variant_eval_step(
        b,
        input_vcf=final_gathered_vcf_job.output_vcf,
        ref_fasta=ref_fasta,
        dbsnp_vcf=dbsnp_vcf,
        output_path=os.path.join(web_bucket, 'variant-eval.txt'),
        disk_size=huge_disk,
    )

    return final_gathered_vcf_job


def add_tabix_step(
    b: hb.Batch,
    vcf_path: str,
    disk_size: int,
) -> Job:
    """
    Regzip and tabix the combined VCF (for some reason the one output with mt2vcf
    is not block-gzipped)
    """
    j = b.new_job('VQSR: Tabix')
    j.image(utils.BCFTOOLS_IMAGE)
    j.memory(f'8G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        combined_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )
    vcf_inp = b.read_input(vcf_path)
    j.command(
        f"""set -e
        gunzip {vcf_inp} -c | bgzip -c > {j.combined_vcf['vcf.gz']}
        tabix -p vcf {j.combined_vcf['vcf.gz']}
        """
    )
    return j


def add_split_intervals_step(
    b: hb.Batch,
    interval_list: hb.ResourceFile,
    scatter_count: int,
    ref_fasta: hb.ResourceGroup,
    disk_size: int,
) -> Job:
    """
    Split genome into intervals to parallelise GnarlyGenotyper.

    Returns: a Job object with a single output j.intervals of type ResourceGroup
    """
    j = b.new_job('VQSR: SplitIntervals')
    j.image(utils.GATK_IMAGE)
    mem_gb = 8
    j.memory(f'{mem_gb}G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        intervals={
            f'interval_{idx}': f'{{root}}/{str(idx).zfill(4)}-scattered.interval_list'
            for idx in range(scatter_count)
        }
    )

    j.command(
        f"""set -e

    # Modes other than INTERVAL_SUBDIVISION will produce an unpredictable number
    # of intervals. But we have to produce exactly {scatter_count} number of
    # output files because our workflow is not dynamic.
    gatk --java-options -Xms{mem_gb - 1}g SplitIntervals \\
      -L {interval_list} \\
      -O {j.intervals} \\
      -scatter {scatter_count} \\
      -R {ref_fasta.base} \\
      -mode INTERVAL_SUBDIVISION
      """
    )
    return j


def add_gnarly_genotyper_on_vcf_step(
    b: hb.Batch,
    combined_gvcf: hb.ResourceGroup,
    ref_fasta: hb.ResourceGroup,
    dbsnp_vcf: hb.ResourceGroup,
    disk_size: int,
    interval: Optional[hb.ResourceGroup] = None,
) -> Job:
    """
    Runs GATK GnarlyGenotyper on a combined_gvcf VCF bgzipped file.

    GnarlyGenotyper performs "quick and dirty" joint genotyping on large cohorts,
    pre-called with HaplotypeCaller, and post-processed with ReblockGVCF.

    HaplotypeCaller must be used with `-ERC GVCF` or `-ERC BP_RESOLUTION` to add
    genotype likelihoods.

    ReblockGVCF must be run to remove low quality variants, as well as to add all the
    annotations necessary for VQSR: QUALapprox, VarDP, RAW_MQandDP.

    Returns: a Job object with a single output j.output_vcf of type ResourceGroup
    """
    j = b.new_job('VQSR: GnarlyGenotyperOnVcf')
    # GnarlyGenotyper crashes with NullPointerException when using standard GATK docker
    j.image(utils.GNARLY_IMAGE)
    j.memory(f'32G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        output_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )

    j.command(
        f"""set -e

    gatk --java-options -Xms8g \\
      GnarlyGenotyper \\
      -R {ref_fasta.base} \\
      -O {j.output_vcf['vcf.gz']} \\
      -D {dbsnp_vcf.base} \\
      --only-output-calls-starting-in-intervals \\
      --keep-all-sites \\
      -V {combined_gvcf['vcf.gz']} \\
      {f'-L {interval} ' if interval else ''} \\
      --create-output-variant-index"""
    )
    return j


def add_hard_filter_step(
    b: hb.Batch,
    input_vcf: hb.ResourceGroup,
    excess_het_threshold: float,
    disk_size: int,
    interval: Optional[hb.ResourceGroup] = None,
) -> Job:
    """
    Hard-filter a large cohort callset on Excess Heterozygosity.

    Applies only to large callsets (`not is_small_callset`)

    Requires all samples to be unrelated.

    ExcessHet estimates the probability of the called samples exhibiting excess
    heterozygosity with respect to the null hypothesis that the samples are unrelated.
    The higher the score, the higher the chance that the variant is a technical artifact
    or that there is consanguinuity among the samples. In contrast to Inbreeding
    Coefficient, there is no minimal number of samples for this annotation.

    Returns: a Job object with a single output j.output_vcf of type ResourceGroup
    """
    j = b.new_job('VQSR: HardFilter')
    j.image(utils.GATK_IMAGE)
    j.memory('8G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        output_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )

    j.command(
        f"""set -euo pipefail

    # Captring stderr to avoid Batch pod from crashing with OOM from millions of
    # warning messages from VariantFiltration, e.g.:
    # > JexlEngine - ![0,9]: 'ExcessHet > 54.69;' undefined variable ExcessHet
    gatk --java-options -Xms3g \\
      VariantFiltration \\
      --filter-expression 'ExcessHet > {excess_het_threshold}' \\
      --filter-name ExcessHet \\
      {f'-L {interval} ' if interval else ''} \\
      -O {j.output_vcf['vcf.gz']} \\
      -V {input_vcf['vcf.gz']} \\
      2> {j.stderr}
    """
    )
    return j


def add_make_sites_only_vcf_step(
    b: hb.Batch,
    input_vcf: hb.ResourceGroup,
    disk_size: int,
    interval: Optional[hb.ResourceGroup] = None,
) -> Job:
    """
    Create sites-only VCF with only site-level annotations.
    Speeds up the analysis in the modeling step.

    Returns: a Job object with a single output j.sites_only_vcf of type ResourceGroup
    """
    j = b.new_job('VQSR: MakeSitesOnlyVcf')
    j.image(utils.GATK_IMAGE)
    j.memory('8G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        sites_only_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )

    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms6g \\
      MakeSitesOnlyVcf \\
      -I {input_vcf['vcf.gz']} \\
      -O {j.sites_only_vcf['vcf.gz']} \\
      {f'-L {interval} ' if interval else ''}"""
    )
    return j


def add_sites_only_gather_vcf_step(
    b: hb.Batch,
    input_vcfs: List[hb.ResourceFile],
    disk_size: int,
) -> Job:
    """
    Gathers VCF files from scattered operations into a single VCF file

    Returns: a Job object with a single output j.output_vcf of type ResourceGroup
    """
    j = b.new_job('VQSR: SitesOnlyGatherVcf')
    j.image(utils.GATK_IMAGE)
    j.memory('8G')
    j.storage(f'{disk_size}G')

    j.declare_resource_group(
        output_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )

    input_cmdl = ' '.join([f'--input {v["vcf.gz"]}' for v in input_vcfs])
    j.command(
        f"""set -euo pipefail

    # --ignore-safety-checks makes a big performance difference so we include it in
    # our invocation. This argument disables expensive checks that the file headers
    # contain the same set of genotyped samples and that files are in order by position
    # of first record.
    gatk --java-options -Xms6g \\
      GatherVcfsCloud \\
      --ignore-safety-checks \\
      --gather-type BLOCK \\
      {input_cmdl} \\
      --output {j.output_vcf['vcf.gz']}

    tabix {j.output_vcf['vcf.gz']}"""
    )
    return j


def add_indels_variant_recalibrator_step(
    b: hb.Batch,
    sites_only_variant_filtered_vcf: hb.ResourceGroup,
    mills_resource_vcf: hb.ResourceGroup,
    axiom_poly_resource_vcf: hb.ResourceGroup,
    dbsnp_resource_vcf: hb.ResourceGroup,
    disk_size: int,
    web_bucket: str = None,
    work_bucket: str = None,
    max_gaussians: int = 4,
) -> Job:
    """
    Run VariantRecalibrator to calculate VQSLOD tranches for indels

    The --max-gaussians parameter sets the expected number of clusters in modeling.
    If a dataset gives fewer distinct clusters, e.g. as can happen for smaller data,
    then the tool will tell you there is insufficient data with a No data found error
    message. In this case, try decrementing the --max-gaussians value. 4 is a
    reasonable default for indels, as their number is smaller than SNPs.

    Returns: a Job object with 3 outputs: j.recalibration (ResourceGroup), j.tranches,
    and j.indel_rscript_file. The latter is usedful to produce the optional tranche plot.
    """
    j = b.new_job('VQSR: IndelsVariantRecalibrator')
    j.image(utils.GATK_IMAGE)
    mem_gb = 32
    j.memory(f'{mem_gb}G')
    j.cpu(2)
    j.storage(f'{disk_size}G')

    j.declare_resource_group(recalibration={'index': '{root}.idx', 'base': '{root}'})

    tranche_cmdl = ' '.join(
        [f'-tranche {v}' for v in INDEL_RECALIBRATION_TRANCHE_VALUES]
    )
    an_cmdl = ' '.join([f'-an {v}' for v in INDEL_RECALIBRATION_ANNOTATION_VALUES])
    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms{mem_gb - 1}g \\
      VariantRecalibrator \\
      -V {sites_only_variant_filtered_vcf['vcf.gz']} \\
      -O {j.recalibration} \\
      --tranches-file {j.tranches} \\
      --trust-all-polymorphic \\
      {tranche_cmdl} \\
      {an_cmdl} \\
      -mode INDEL \\
      --use-allele-specific-annotations \\
      --max-gaussians {max_gaussians} \\
      -resource:mills,known=false,training=true,truth=true,prior=12 {mills_resource_vcf.base} \\
      -resource:axiomPoly,known=false,training=true,truth=false,prior=10 {axiom_poly_resource_vcf.base} \\
      -resource:dbsnp,known=true,training=false,truth=false,prior=2 {dbsnp_resource_vcf.base} \\
      --rscript-file {j.indel_rscript_file}
      
      ls $(dirname {j.indel_rscript_file})

      ln {j.indel_rscript_file}.pdf {j.indel_features_pdf}
      """
    )
    if work_bucket:
        b.write_output(
            j.indel_rscript_file,
            os.path.join(work_bucket, 'recalibration-indels-features.Rscript'),
        )
    if web_bucket:
        b.write_output(
            j.indel_features_pdf,
            os.path.join(web_bucket, 'recalibration-indels-features.pdf'),
        )
    return j


def add_snps_variant_recalibrator_create_model_step(
    b: hb.Batch,
    sites_only_variant_filtered_vcf: hb.ResourceGroup,
    hapmap_resource_vcf: hb.ResourceGroup,
    omni_resource_vcf: hb.ResourceGroup,
    one_thousand_genomes_resource_vcf: hb.ResourceGroup,
    dbsnp_resource_vcf: hb.ResourceGroup,
    disk_size: int,
    web_bucket: str = None,
    work_bucket: str = None,
    is_small_callset: bool = False,
    is_huge_callset: bool = False,
    max_gaussians: int = 4,
) -> Job:
    """
    First step of VQSR for SNPs: run VariantRecalibrator to subsample variants
    and produce a file of the VQSR model.

    To support cohorts with more than 10,000 WGS samples, the SNP recalibrartion process
    is borken down across genomic regions for parallel processing, and done in 3 steps:
    1. Run the recalibrator with the following additional arguments:
       --sample-every-Nth-variant <downsample_factor> --output-model <model_file>
    2. Apply the resulting model to each genomic interval with, running the recalibrator
       with the same base parameters, plus:
       --input-model <model-file> --output-tranches-for-scatter
    3. Collate the resulting per-interval tranches with GatherTranches

    The --max-gaussians parameter sets the expected number of clusters in modeling.
    If a dataset gives fewer distinct clusters, e.g. as can happen for smaller data,
    then the tool will tell you there is insufficient data with a No data found error
    message. In this case, try decrementing the --max-gaussians value.

    Returns: a Job object with 2 outputs: j.model and j.snp_rscript_file.
    The latter is useful to produce the optional tranche plot.
    """
    j = b.new_job('VQSR: SNPsVariantRecalibratorCreateModel')
    j.image(utils.GATK_IMAGE)
    mem_gb = 64 if not is_small_callset else 128
    j.memory(f'{mem_gb}G')
    j.cpu(2)
    j.storage(f'{disk_size}G')

    downsample_factor = 75 if is_huge_callset else 10

    tranche_cmdl = ' '.join([f'-tranche {v}' for v in SNP_RECALIBRATION_TRANCHE_VALUES])
    an_cmdl = ' '.join([f'-an {v}' for v in SNP_RECALIBRATION_ANNOTATION_VALUES])
    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms{mem_gb - 2}g \\
      VariantRecalibrator \\
      -V {sites_only_variant_filtered_vcf['vcf.gz']} \\
      -O {j.recalibration} \\
      --tranches-file {j.tranches} \\
      --trust-all-polymorphic \\
      {tranche_cmdl} \\
      {an_cmdl} \\
      -mode SNP \\
      --use-allele-specific-annotations \\
      --sample-every-Nth-variant {downsample_factor} \\
      --output-model {j.model_file} \\
      --max-gaussians {max_gaussians} \\
      -resource:hapmap,known=false,training=true,truth=true,prior=15 {hapmap_resource_vcf.base} \\
      -resource:omni,known=false,training=true,truth=true,prior=12 {omni_resource_vcf.base} \\
      -resource:1000G,known=false,training=true,truth=false,prior=10 {one_thousand_genomes_resource_vcf.base} \\
      -resource:dbsnp,known=true,training=false,truth=false,prior=7 {dbsnp_resource_vcf.base} \\
      --rscript-file {j.snp_rscript}

      ls $(dirname {j.snp_rscript})

      ln {j.snp_rscript}.pdf {j.snp_rscript_pdf}
      ln {j.tranches}.pdf {j.tranches_pdf}
      """
    )
    if work_bucket:
        b.write_output(
            j.snp_rscript,
            os.path.join(work_bucket, 'recalibration-snps-features.RScript'),
        )
    if web_bucket:
        b.write_output(
            j.snp_rscript_pdf,
            os.path.join(web_bucket, 'recalibration-snps-features.pdf'),
        )
        b.write_output(
            j.tranches_pdf,
            os.path.join(web_bucket, 'recalibration-snps-tranches.pdf'),
        )
    return j


def add_snps_variant_recalibrator_scattered_step(
    b: hb.Batch,
    sites_only_vcf: hb.ResourceGroup,
    model_file: hb.ResourceGroup,
    hapmap_resource_vcf: hb.ResourceGroup,
    omni_resource_vcf: hb.ResourceGroup,
    one_thousand_genomes_resource_vcf: hb.ResourceGroup,
    dbsnp_resource_vcf: hb.ResourceGroup,
    disk_size: int,
    interval: Optional[hb.ResourceGroup] = None,
    max_gaussians: int = 4,
) -> Job:
    """
    Second step of VQSR for SNPs: run VariantRecalibrator scattered to apply
    the VQSR model file to each genomic interval.

    To support cohorts with more than 10,000 WGS samples, the SNP recalibrartion process
    is borken down across genomic regions for parallel processing, and done in 3 steps:
    1. Run the recalibrator with the following additional arguments:
       --sample-every-Nth-variant <downsample_factor> --output-model <model_file>
    2. Apply the resulting model to each genomic interval with, running the recalibrator
       with the same base parameters, plus:
       --input-model <model-file> --output-tranches-for-scatter
    3. Collate the resulting per-interval tranches with GatherTranches

    The --max-gaussians parameter sets the expected number of clusters in modeling.
    If a dataset gives fewer distinct clusters, e.g. as can happen for smaller data,
    then the tool will tell you there is insufficient data with a No data found error
    message. In this case, try decrementing the --max-gaussians value.

    Returns: a Job object with 2 outputs: j.recalibration (ResourceGroup) and j.tranches
    """
    j = b.new_job('VQSR: SNPsVariantRecalibratorScattered')

    j.image(utils.GATK_IMAGE)
    mem_gb = 64  # ~ twice the sum of all input resources and input VCF sizes
    j.memory(f'{mem_gb}G')
    j.cpu(2)
    j.storage(f'{disk_size}G')

    j.declare_resource_group(recalibration={'index': '{root}.idx', 'base': '{root}'})

    tranche_cmdl = ' '.join([f'-tranche {v}' for v in SNP_RECALIBRATION_TRANCHE_VALUES])
    an_cmdl = ' '.join([f'-an {v}' for v in SNP_RECALIBRATION_ANNOTATION_VALUES])
    j.command(
        f"""set -euo pipefail

    MODEL_REPORT={model_file}

    gatk --java-options -Xms{mem_gb - 1}g \\
      VariantRecalibrator \\
      -V {sites_only_vcf['vcf.gz']} \\
      -O {j.recalibration} \\
      --tranches-file {j.tranches} \\
      --trust-all-polymorphic \\
      {tranche_cmdl} \\
      {an_cmdl} \\
      -mode SNP \\
      {f'-L {interval} ' if interval else ''} \\
      --use-allele-specific-annotations \\
      --input-model {model_file} --output-tranches-for-scatter \\
      --max-gaussians {max_gaussians} \\
      -resource:hapmap,known=false,training=true,truth=true,prior=15 {hapmap_resource_vcf.base} \\
      -resource:omni,known=false,training=true,truth=true,prior=12 {omni_resource_vcf.base} \\
      -resource:1000G,known=false,training=true,truth=false,prior=10 {one_thousand_genomes_resource_vcf.base} \\
      -resource:dbsnp,known=true,training=false,truth=false,prior=7 {dbsnp_resource_vcf.base}"""
    )
    return j


def add_snps_variant_recalibrator_step(
    b: hb.Batch,
    sites_only_variant_filtered_vcf: hb.ResourceGroup,
    hapmap_resource_vcf: hb.ResourceGroup,
    omni_resource_vcf: hb.ResourceGroup,
    one_thousand_genomes_resource_vcf: hb.ResourceGroup,
    dbsnp_resource_vcf: hb.ResourceGroup,
    web_bucket: str,
    work_bucket: str,
    disk_size: int,
    max_gaussians: int = 4,
) -> Job:
    """
    Recalibrate SNPs in one run (alternative to scatter-gather approach)
    """
    j = b.new_job('VQSR: SNPsVariantRecalibrator')

    j.image(utils.GATK_IMAGE)
    mem_gb = 64  # ~ twice the sum of all input resources and input VCF sizes
    j.memory(f'{mem_gb}G')
    j.cpu(2)
    j.storage(f'{disk_size}G')

    j.declare_resource_group(recalibration={'index': '{root}.idx', 'base': '{root}'})

    tranche_cmdl = ' '.join([f'-tranche {v}' for v in SNP_RECALIBRATION_TRANCHE_VALUES])
    an_cmdl = ' '.join([f'-an {v}' for v in SNP_RECALIBRATION_ANNOTATION_VALUES])
    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms{mem_gb - 1}g \\
      VariantRecalibrator \\
      -V {sites_only_variant_filtered_vcf['vcf.gz']} \\
      -O {j.recalibration} \\
      --tranches-file {j.tranches} \\
      --trust-all-polymorphic \\
      {tranche_cmdl} \\
      {an_cmdl} \\
      -mode SNP \\
      --use-allele-specific-annotations \\
      --max-gaussians {max_gaussians} \\
      -resource:hapmap,known=false,training=true,truth=true,prior=15 {hapmap_resource_vcf.base} \\
      -resource:omni,known=false,training=true,truth=true,prior=12 {omni_resource_vcf.base} \\
      -resource:1000G,known=false,training=true,truth=false,prior=10 {one_thousand_genomes_resource_vcf.base} \\
      -resource:dbsnp,known=true,training=false,truth=false,prior=7 {dbsnp_resource_vcf.base} \\
      --rscript-file {j.snp_rscript}

      ln {j.snp_rscript}.pdf {j.snp_rscript_pdf}
      ln {j.tranches}.pdf {j.tranches_pdf}
      """
    )

    if work_bucket:
        b.write_output(
            j.snp_rscript,
            os.path.join(work_bucket, 'recalibration-snps-features.RScript'),
        )
    if web_bucket:
        b.write_output(
            j.snp_rscript_pdf,
            os.path.join(web_bucket, 'recalibration-snps-features.pdf'),
        )
        b.write_output(
            j.tranches_pdf,
            os.path.join(web_bucket, 'recalibration-snps-tranches.pdf'),
        )
    return j


def add_snps_gather_tranches_step(
    b: hb.Batch,
    tranches: List[hb.ResourceFile],
    disk_size: int,
) -> Job:
    """
    Third step of VQSR for SNPs: run GatherTranches to gather scattered per-interval
    tranches outputs.

    To support cohorts with more than 10,000 WGS samples, the SNP recalibrartion process
    is borken down across genomic regions for parallel processing, and done in 3 steps:
    1. Run the recalibrator with the following additional arguments:
       --sample-every-Nth-variant <downsample_factor> --output-model <model_file>
    2. Apply the resulting model to each genomic interval with, running the recalibrator
       with the same base parameters, plus:
       --input-model <model-file> --output-tranches-for-scatter
    3. Collate the resulting per-interval tranches with GatherTranches

    Returns: a Job object with one output j.out_tranches
    """
    j = b.new_job('VQSR: SNPGatherTranches')
    j.image(utils.GATK_IMAGE)
    j.memory('8G')
    j.cpu(2)
    j.storage(f'{disk_size}G')

    inputs_cmdl = ' '.join([f'--input {t}' for t in tranches])
    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms6g \\
      GatherTranches \\
      --mode SNP \\
      {inputs_cmdl} \\
      --output {j.out_tranches}"""
    )
    return j


def add_apply_recalibration_step(
    b: hb.Batch,
    input_vcf: hb.ResourceFile,
    indels_recalibration: hb.ResourceGroup,
    indels_tranches: hb.ResourceFile,
    snps_recalibration: hb.ResourceGroup,
    snps_tranches: hb.ResourceFile,
    disk_size: int,
    indel_filter_level: float,
    snp_filter_level: float,
    interval: Optional[hb.ResourceGroup] = None,
    output_vcf_path: Optional[str] = None,
) -> Job:
    """
    Apply a score cutoff to filter variants based on a recalibration table.
    Runs ApplyVQSR twice to apply first indel, then SNP recalibrations.

    Targets indel_filter_level and snp_filter_level sensitivities. The tool matches
    them internally to a VQSLOD score cutoff based on the model's estimated sensitivity
    to a set of true variants.

    The filter determination is not just a pass/fail process. The tool evaluates for
    each variant which "tranche", or slice of the dataset, it falls into in terms of
    sensitivity to the truthset. Variants in tranches that fall below the specified
    truth sensitivity filter level have their FILTER field annotated with the
    corresponding tranche level. This results in a callset that is filtered to the
    desired level but retains the information necessary to increase sensitivity
    if needed.

    Returns: a Job object with one ResourceGroup output j.recalibrated_vcf, correponding
    to a VCF with tranche annotated in the FILTER field
    """
    j = b.new_job('VQSR: ApplyRecalibration')
    j.image(utils.GATK_IMAGE)
    j.memory('8G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        recalibrated_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )

    j.command(
        f"""set -euo pipefail
        
    df -h; pwd; du -sh $(dirname {j.recalibrated_vcf['vcf.gz']})
    
    TMP_DIR=$(dirname {j.recalibrated_vcf['vcf.gz']})/tmp
    mkdir $TMP_DIR

    gatk --java-options -Xms5g \\
      ApplyVQSR \\
      --tmp-dir $TMP_DIR \\
      -O tmp.indel.recalibrated.vcf.gz \\
      -V {input_vcf['vcf.gz']} \\
      --recal-file {indels_recalibration} \\
      --tranches-file {indels_tranches} \\
      --truth-sensitivity-filter-level {indel_filter_level} \\
      --create-output-variant-index true \\
      -mode INDEL \\
      {f'-L {interval} ' if interval else ''} \\
      --use-allele-specific-annotations

    df -h; pwd; du -sh $(dirname {j.recalibrated_vcf['vcf.gz']})

    rm {input_vcf['vcf.gz']} {indels_recalibration} {indels_tranches}
    rm -rf $TMP_DIR
    mkdir $TMP_DIR

    df -h; pwd; du -sh $(dirname {j.recalibrated_vcf['vcf.gz']})

    gatk --java-options -Xms5g \\
      ApplyVQSR \\
      --tmp-dir $TMP_DIR \\
      -O {j.recalibrated_vcf['vcf.gz']} \\
      -V tmp.indel.recalibrated.vcf.gz \\
      --recal-file {snps_recalibration} \\
      --tranches-file {snps_tranches} \\
      --truth-sensitivity-filter-level {snp_filter_level} \\
      --create-output-variant-index true \\
      -mode SNP \\
      {f'-L {interval} ' if interval else ''} \\
      --use-allele-specific-annotations

    df -h; pwd; du -sh $(dirname {j.recalibrated_vcf['vcf.gz']})
    """
    )

    if output_vcf_path:
        b.write_output(j.recalibrated_vcf, output_vcf_path.replace('.vcf.gz', ''))
    return j


def add_collect_metrics_sharded_step(
    b: hb.Batch,
    input_vcf: hb.ResourceGroup,
    dbsnp_vcf: hb.ResourceGroup,
    interval_list: hb.ResourceFile,
    ref_dict: hb.ResourceFile,
    disk_size: int,
):
    """
    Run CollectVariantCallingMetrics for site-level evaluation.

    This produces detailed and summary metrics report files. The summary metrics
    provide cohort-level variant metrics and the detailed metrics segment variant
    metrics for each sample in the callset. The detail metrics give the same metrics
    as the summary metrics for the samples plus several additional metrics.

    These are explained in detail at
    https://broadinstitute.github.io/picard/picard-metric-definitions.html.

    Returns: Job object with a single ResourceGroup output j.metrics, with
    j.metrics.detail_metrics and j.metrics.summary_metrics ResourceFiles
    """
    j = b.new_job('VQSR: CollectMetricsSharded')
    j.image(utils.GATK_IMAGE)
    j.memory('8G')
    j.cpu(2)
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        metrics={
            'detail_metrics': '{root}.variant_calling_detail_metrics',
            'summary_metrics': '{root}.variant_calling_summary_metrics',
        }
    )

    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms6g \\
      CollectVariantCallingMetrics \\
      --INPUT {input_vcf['vcf.gz']} \\
      --DBSNP {dbsnp_vcf.base} \\
      --SEQUENCE_DICTIONARY {ref_dict} \\
      --OUTPUT {j.metrics} \\
      --THREAD_COUNT 8 \\
      --TARGET_INTERVALS {interval_list}"""
    )
    return j


def _add_final_gather_vcf_step(
    b: hb.Batch,
    input_vcfs: List[hb.ResourceGroup],
    disk_size: int,
    output_vcf_path: str = None,
) -> Job:
    """
    Combines recalibrated VCFs into a single VCF.
    Saves the output VCF to a bucket `output_vcf_path`
    """
    j = b.new_job('VQSR: FinalGatherVcf')
    j.image(utils.GATK_IMAGE)
    j.memory(f'8G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        output_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )

    input_cmdl = ' '.join([f'--input {v["vcf.gz"]}' for v in input_vcfs])
    j.command(
        f"""set -euo pipefail

    # --ignore-safety-checks makes a big performance difference so we include it in 
    # our invocation. This argument disables expensive checks that the file headers 
    # contain the same set of genotyped samples and that files are in order 
    # by position of first record.
    gatk --java-options -Xms6g \\
      GatherVcfsCloud \\
      --ignore-safety-checks \\
      --gather-type BLOCK \\
      {input_cmdl} \\
      --output {j.output_vcf['vcf.gz']}

    tabix {j.output_vcf['vcf.gz']}"""
    )
    if output_vcf_path:
        b.write_output(j.output_vcf, output_vcf_path.replace('.vcf.gz', ''))
    return j


def _add_filter_sb_step(
    b: hb.Batch,
    input_vcf: hb.ResourceGroup,
    disk_size: int,
    output_vcf_path: str = None,
) -> Job:
    """
    Removes the INFO/SB field from a VCF.

    The reason we are doing that is because gatk ApplyVQSR replaces the VCF header
    ##INFO=<ID=SB,Number=.,Type=Int,Description="Strand Bias">

    with
    ##INFO=<ID=SB,Number=1,Type=Float,Description="Strand Bias">

    Even though the actual SB field is still a list of integers: SB=5,2,18,29
    It breaks parsing the VCF into Hail.
    """
    j = b.new_job('VQSR: Remove SB')
    j.image(utils.BCFTOOLS_IMAGE)
    j.memory(f'8G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        output_vcf={'vcf.gz': '{root}.vcf.gz', 'vcf.gz.tbi': '{root}.vcf.gz.tbi'}
    )

    j.command(
        f"""
    bcftools annotate -x INFO/SB {input_vcf['vcf.gz']} -Oz -o {j.output_vcf['vcf.gz']} && tabix {j.output_vcf['vcf.gz']}
    """
    )
    if output_vcf_path:
        b.write_output(j.output_vcf, output_vcf_path.replace('.vcf.gz', ''))
    return j


def _add_variant_eval_step(
    b: hb.Batch,
    input_vcf: hb.ResourceGroup,
    ref_fasta: hb.ResourceGroup,
    dbsnp_vcf: hb.ResourceGroup,
    disk_size: int,
    output_path: str = None,
) -> Job:
    """
    Run VariantEval for site-level evaluation.
    Saves the QC to `output_path` bucket
    """
    j = b.new_job('VQSR: VariantEval')
    j.image(utils.GATK_IMAGE)
    j.memory(f'8G')
    j.storage(f'{disk_size}G')

    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms6g \\
      VariantEval \\
      --eval {input_vcf['vcf.gz']} \\
      -R {ref_fasta.base} \\
      -D {dbsnp_vcf.base} \\
      --output {j.output}"""
    )
    if output_path:
        b.write_output(j.output, output_path)
    return j


def add_gather_variant_calling_metrics_step(
    b: hb.Batch,
    input_details: List[hb.ResourceGroup],
    input_summaries: List[hb.ResourceGroup],
    disk_size: int,
    output_path_prefix: str = None,
) -> Job:
    """
    Combines metrics from multiple CollectVariantCallingMetrics runs.

    Returns: Job object with a single ResourceGroup output j.metrics, with
    j.metrics.detail_metrics and j.metrics.summary_metrics ResourceFiles

    Saves the QC results to a bucket with the `output_path_prefix` prefix
    """
    j = b.new_job('VQSR: GatherVariantCallingMetrics')
    j.image(utils.GATK_IMAGE)
    j.memory(f'8G')
    j.storage(f'{disk_size}G')
    j.declare_resource_group(
        metrics={
            'detail_metrics': '{root}.variant_calling_detail_metrics',
            'summary_metrics': '{root}.variant_calling_summary_metrics',
        }
    )

    input_cmdl = ' '.join('--INPUT {f} ' for f in input_details + input_summaries)
    j.command(
        f"""set -euo pipefail

    gatk --java-options -Xms2g \\
      AccumulateVariantCallingMetrics \\
      {input_cmdl} \\
      --OUTPUT {j.metrics}"""
    )
    if output_path_prefix:
        b.write_output(j.metrics, output_path_prefix)
    return j
